/* create-rewrite-read tests.
   Author:  Raphael Finkel, 7/1997
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

/* constants */
#define FILENAMESIZE 100

/* test parameters */
int Increment;
int MaxFileSize;
int NumFiles;
char Path[50];

/* initialized constants */
char *DataBuffer;

/* variables */
int FileSize;

/* timing macros */

double elapsed;

#include <sys/time.h>
struct  timeval TV1, TV2;
struct timezone TZ;
#define TIME(which) gettimeofday(&which, &TZ)
#define DIFF(meas1, meas2)  /* in microseconds */ \
	(((meas2.tv_sec - meas1.tv_sec) * 1000000L) +  \
	(meas2.tv_usec - meas1.tv_usec))
 
#define BEGINTEST(message) \
	TIME(TV1)
#define ENDTEST(message) \
	TIME(TV2); \
	fprintf(stdout, "\t%s\t%g sec\n", message, \
		DIFF(TV1, TV2) / 1000000.0); \
	fflush(stdout);
#	define DOTEST(message, code) \
	BEGINTEST(message); code; ENDTEST(message);

void Init() {
	int *DataPtr;
	int Count;
	int NumInts = MaxFileSize / sizeof(int);
	DataBuffer = malloc(2*MaxFileSize);
	if (!DataBuffer) {
		fprintf(stderr, "Out of memory");
	}
	/* touch every word of DataBuffer to force into memory if possible */
	DataPtr = (int *) DataBuffer;
	for (Count = 0; Count < NumInts; Count++) {
		*DataPtr++ = Count;
	}
} /* Init */

void CreateTest() {
	int FileNumber, FileDescriptor, Result;
	char FileName[FILENAMESIZE];
	for (FileNumber = 1; FileNumber <= NumFiles; FileNumber++) {
		sprintf(FileName, "%sfile.%d", Path, FileNumber);
		FileDescriptor = creat(FileName, 0666);
		if (FileDescriptor < 0) {
			perror(FileName);
			exit(1);
		}
		Result = write(FileDescriptor, DataBuffer, FileSize);
		if (Result != FileSize) {
			perror(FileName);
			exit(1);
		}
		Result = close(FileDescriptor);
		if (Result) {
			perror(FileName);
			exit(1);
		}
	} /* for each file */
} /* CreateTest */

void RewriteTest() {
	int FileNumber, FileDescriptor, Result;
	char FileName[FILENAMESIZE];
	for (FileNumber = 1; FileNumber <= NumFiles; FileNumber++) {
		sprintf(FileName, "%sfile.%d", Path, FileNumber);
		FileDescriptor = open(FileName, O_WRONLY);
		if (FileDescriptor < 0) {
			perror(FileName);
			exit(1);
		}
		Result = write(FileDescriptor, DataBuffer, FileSize);
		if (Result != FileSize) {
			perror(FileName);
			exit(1);
		}
		Result = close(FileDescriptor);
		if (Result) {
			perror(FileName);
			exit(1);
		}
	} /* for each file */
} /* RewriteTest */

void ReadTest() {
	int FileNumber, FileDescriptor, Result;
	char FileName[FILENAMESIZE];
	for (FileNumber = 1; FileNumber <= NumFiles; FileNumber++) {
		sprintf(FileName, "%sfile.%d", Path, FileNumber);
		FileDescriptor = open(FileName, O_RDONLY);
		if (FileDescriptor < 0) {
			perror(FileName);
			exit(1);
		}
		Result = read(FileDescriptor, DataBuffer, FileSize);
		if (Result != FileSize) {
			perror(FileName);
			exit(1);
		}
		Result = close(FileDescriptor);
		if (Result) {
			perror(FileName);
			exit(1);
		}
	} /* for each file */
} /* ReadTest */

void DeleteTest() {
	int FileNumber, Result;
	char FileName[FILENAMESIZE];
	for (FileNumber = 1; FileNumber <= NumFiles; FileNumber++) {
		sprintf(FileName, "%sfile.%d", Path, FileNumber);
		Result = unlink(FileName);
		if (Result) {
			perror(FileName);
			exit(1);
		}
	} /* for each file */
} /* DeleteTest */

int main(int argc, char *argv[]) {
	if (argc != 5) {
		fprintf(stderr, "Usage: %s DirPath, NumFiles, Increment (in K), "
						"NumIncrements\n", argv[0]);
		return(1);
	}
	strcpy(Path,argv[1]);
	NumFiles = atoi(argv[2]);
	Increment = atoi(argv[3]) * 1024 * 1024;
	MaxFileSize = atoi(argv[4]) * Increment;
	Init();
	for (FileSize = 0; FileSize <= MaxFileSize; FileSize += Increment) {
		fprintf(stdout, "File size %d; %d files\n", FileSize, NumFiles);
		DOTEST("Create", CreateTest());
		DOTEST("Sync", sync());
		DOTEST("Rewrite", RewriteTest());
		DOTEST("Sync", sync());
		DOTEST("Read", ReadTest());
		DOTEST("Sync", sync());
		DOTEST("Delete", DeleteTest());
		DOTEST("Sync", sync());
	}
	return(0);
}
