/* -----------------------------------------------------------------
The VIVA file system - version 2.0

Implemented by Shankar Pasupathy, July 1996. (shankar@pop.uky.edu)
Modified by Vamsi Krishna, July 1997. (vamsi@cslab.uky.edu)
VIVA was developed by Eric Herrin II and Raphael Finkel.
Department of Computer Science,
University of Kentucky, Lexington 40506.
Distribution site: http://www.cslab.uky.edu/~vamsi/viva

This implementation uses code from the Minix filesystem,
(Copyright: Linus Torvalds).

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation. Read the file COPYRIGHT, included with
this distribution, for further details.
--------------------------------------------------------------------*/

#ifndef _VIVA_FS_SB
#define _VIVA_FS_SB

#include <linux/viva_fs.h>

/*
 * viva super-block data in memory
 */
struct viva_sb_info {
	unsigned int s_dev;
	unsigned long s_ninodes; /* total number of inodes */
	unsigned long s_nzones; /* total number of zones */
	unsigned long s_imap_blocks; /* blocks used for inode map */
	unsigned long s_zmap_blocks; /* blocks used for zone map */
	unsigned long s_fsmap_blocks; /* blocks used for frag_sparse map */
	unsigned long s_famap_blocks; /* blocks used for frag_alloc map */
	unsigned long s_firstdatazone; /* first data block */
	unsigned long s_first_fsmapblock; /* first fsmap block */
	unsigned long s_frag_blocks; /* Number of Fragment blocks allocated */
	unsigned long s_max_frag_blocks; /* Max number of Frag blocks allowed */
	unsigned long s_last_frag_block; /* last allocated fragment block */
	unsigned long s_log_zone_size; /* not used in this version */
	unsigned short s_inode_nsegments;
	unsigned short s_zone_nsegments;
	unsigned short s_inode_seg_size;
	unsigned short s_zone_seg_size;
	unsigned short s_inode_lastseg_size;
	unsigned short s_zone_lastseg_size;
	struct buffer_head * s_imap[VIVA_I_MAP_SLOTS]; /* inode bitmap */
	struct buffer_head * s_zmap[VIVA_Z_MAP_SLOTS]; /* zone bitmap */
	struct buffer_head * s_famap[VIVA_FA_MAP_SLOTS]; /* fragment alloc bitmap */
	unsigned char s_cached_fsmap[VIVA_FS_MAP_SIZE];
	unsigned long s_free_inodes[4][VIVA_MAX_INODE_SEGMENTS]; /* inode segments*/
	unsigned long s_free_zones[4][VIVA_MAX_ZONE_SEGMENTS]; /* zone segments */
	unsigned long s_dirsize; /* size of a VIVA directory */
	unsigned long s_namelen; /* max size of a file name */
	struct buffer_head * s_sbh; /* pointer to the buffer that 
								will hold the super-block */
	struct viva_super_block * s_vs; /* pointer to the VIVA 
									super-block structure */
	unsigned short s_mount_state; /* valid fs or error fs */
};

#endif
