/* -----------------------------------------------------------------
The VIVA file system - version 2.0

Implemented by Shankar Pasupathy, July 1996. (shankar@pop.uky.edu)
Modified by Vamsi Krishna, July 1997. (vamsi@cslab.uky.edu)
VIVA was developed by Eric Herrin II and Raphael Finkel.
Department of Computer Science,
University of Kentucky, Lexington 40506.
Distribution site: http://www.cslab.uky.edu/~vamsi/viva

This implementation uses code from the Minix filesystem,
(Copyright: Linus Torvalds).

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation. Read the file COPYRIGHT, included with
this distribution, for further details.
--------------------------------------------------------------------*/
/* bits_set array values - thanks to Eric Herrin */
/* These functions are the core of VIVA. viva_bmap was very difficult
to implement */


#include <linux/sched.h>
#include <linux/viva_fs.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/locks.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/bitops.h>

#define count_bits_set(byte) (bits_set[*(unsigned char *)(byte)])

unsigned char bits_set[256] = { /* number of 1s in a byte */
    0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4,
    1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
    1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
    1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
    2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
    3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
    3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
    4,5,5,6,5,6,6,7,5,6,6,7,6,7,7,8
};

/* --------------------------------------------------------------
viva_bmap: Returns the disk block address of a block number for 
a given inode. For example, viva_bmap(sparse_bmap, 1) returns the 
address of block number 1 in "inode". The function returns 0 on
error.

This function performs poorly when mapping large numbers to addresses
i.e. when the inode refers to a large file. We plan to store the 
last block number mapped, in memory, and use this to speed up
viva_bmap.
-----------------------------------------------------------------*/

int viva_bmap(unsigned char *sparse_bmap, int nr)
{
int i,block_nr,bit_nr,byte; /* nr stands for number */
unsigned char temp_byte[1];
struct viva_smap_entry *entry; /* defined in viva_fs.h */
int count;

if (nr<0) {
	printk("VIVA: In viva_bmap:block number<0");
	return 0;
}

/* Never test for nr > blocksInUse - it causes a horrible bug in truncate.
truncate starts to free block 0, block 1 and so on. But blocksInUse will soon
become < block_nr in truncate causing only half the blocks in a file to be
freed */

block_nr = -1;

entry = (struct viva_smap_entry *) (sparse_bmap);

while (entry->start_addr || entry->length) {
	if(entry->start_addr)
	{
		block_nr++; /* because the first bit is implicitly set */
		/* and for the same reason check the following: */
		if (nr == block_nr) 
			return(entry->start_addr);
		byte = 0;
		while (byte < entry->length) 
		{
			if (nr < (block_nr + count_bits_set(entry->bitmap + byte)+1)) 
				goto block_in_this_byte;
			block_nr += count_bits_set(entry->bitmap + byte);
			byte++;
		}
		entry = (struct viva_smap_entry *) (entry->bitmap + entry->length);
	}
	else
	{
		count = entry->length;
		while(count)
		{
			block_nr++;
			if (nr == block_nr) 
			{
/*				printk(" Trouble %d (%d)",nr,count);*/
				return -1;  /* I think I need to change the return value */
			}
			entry = (struct viva_smap_entry *) (entry->bitmap + 0);
			count--;
		}
	}
}
/* block requested in not available */
return 0;

block_in_this_byte:
	memcpy(temp_byte,(entry->bitmap+byte),1);
	*temp_byte = *temp_byte ^ 0xff; /* make all 0s, 1s */
	bit_nr = -1;
	for (i=0; i<(nr-block_nr); i++) {
		bit_nr = find_next_zero_bit(temp_byte,8,bit_nr+1);
	}
	if (bit_nr > 7) {
		printk("VIVA: In viva_bmap: Error in sparse bitmap\n");
/*		printk("Vamsi: bit_nr = %d, nr = %d, block_nr = %d, temp_byte = %d\n",bit_nr,nr,block_nr,temp_byte);*/
		return 0; /* error */
	}
	return(entry->start_addr + byte * 8 + bit_nr + 1);
}

int viva_break_hole(struct inode * inode, int new_block_no, int new_block_addr)
{
	struct viva_smap_entry *entry;
	int temp, count, byte, block_nr;
	unsigned long first_valid_addr;
	unsigned long last_valid_addr;

	first_valid_addr = (unsigned long)(entry);

	entry = (struct viva_smap_entry *) (inode->u.viva_i.i_cached_smap);
	block_nr = 0;
	while (entry->start_addr || entry->length) 
	{
		if(entry->start_addr)
		{
			last_valid_addr = entry->start_addr;
			block_nr++;
			byte = 0;
			while (byte < entry->length) 
			{
				block_nr += count_bits_set(entry->bitmap + byte);
				byte++;
			}
			entry = (struct viva_smap_entry *) (entry->bitmap + entry->length);
		}
		else
		{
			count = entry->length;
			if( (new_block_no >= block_nr) && 
					(new_block_no < block_nr+count) )
			{
/*				printk("break_hole: First part of the hole %d %d %d\n", new_block_no, block_nr, count);*/

				/* First part of the hole */
				entry->length = new_block_no - block_nr;
				for(temp = 1; temp < entry->length; temp++)
				{
					entry = (struct viva_smap_entry *) (entry->bitmap + 0);
				}

				/* Second part of the hole, which has the new inserted block */
				entry->start_addr = new_block_addr;
				entry->length = 0;
				entry = (struct viva_smap_entry *) (entry->bitmap + 0);

				/* Third part of the hole, only if necessary */
				if( (block_nr+count) - new_block_no - 1 )
				{
					entry->start_addr = 0;
					entry->length = (block_nr+count) - new_block_no - 1;
				}
				block_nr += count;
				for(temp = 1; temp <= entry->length; temp++)
				{
					entry = (struct viva_smap_entry *) (entry->bitmap + 0);
				}
/*				printk("break_hole: %d %d Done \n",entry->start_addr, entry->length);*/
				return(0);
			}
			else
			{
				block_nr += count;
				for(temp = 1; temp <= count; temp++)
					entry = (struct viva_smap_entry *)(entry->bitmap + 0);
			}
		}
	}
	return(-1);
}
	

int viva_add_hole(struct inode * inode, int num_holes)
{

	struct viva_smap_entry *entry;
	unsigned long last_valid_addr;
	int count;

	entry = (struct viva_smap_entry *) (inode->u.viva_i.i_cached_smap);
	last_valid_addr = (unsigned long)(entry) + VIVA_SMAP_SIZE - 1;
	if (entry->start_addr) 
	{
		while (entry->start_addr || entry->length) 
		{
			if(entry->start_addr)
				entry = (struct viva_smap_entry *)
								(entry->bitmap + entry->length);
			else
			{
				count = entry->length;
				while(count)
				{
					entry = (struct viva_smap_entry *)(entry->bitmap + 0);
					count--;
				}
			}
		}
	}
	if ((unsigned long)entry > last_valid_addr) 
	{
		return -2; /* ran out of space */
	}
	entry->start_addr = 0;
	entry->length = num_holes;
	entry = (struct viva_smap_entry *)(entry->bitmap + 0);
	for(count=1; count < num_holes; count++)
	{
		entry->start_addr = 0;
		entry->length = 0;
		entry = (struct viva_smap_entry *)(entry->bitmap + 0);
	}
	inode->u.viva_i.i_blocksInUse += num_holes;
	inode->i_ctime = CURRENT_TIME;
	inode->i_dirt = 1;
	return 1;
}


/* -----------------------------------------------------------------------
viva_expand_smap: Another horrible function to implement. It expands a
sparse bitmap to indicate a new block. There are plenty of boundary 
conditions to check for in this function so don't modify it unless you
understand VIVA's design completely. 
-------------------------------------------------------------------------- */

int viva_expand_smap(unsigned char *sparse_bmap, int new_block_addr, 
					 int smap_size )
{
/* Bug that was fixed: If the length of the bitmap exceeds 255
bytes, we must create a new entry since the entry length is an unsigned char
*/
/* Algorithm: 
Wade through the sparse bitmap till we find the last bit that is set. 
Compute the corresponding block's address.
If (new_block_addr - last_block_addr) < 32, set the proper bit in the
current entry. Update the length of the current entry.
Otherwise, create a new entry with the starting address set to new_block_addr
and entry_length set to zero 
*/

/* CHECK if bit to be set is greater than there is space in the sparse
bitmap */

struct viva_smap_entry *entry,*next_entry;
char last_byte[1];
unsigned long last_block_addr,last_valid_addr,first_valid_addr;
int old_bit_nr, bit_nr, bit_to_set;
int count;

entry = next_entry = (struct viva_smap_entry *)(sparse_bmap);
last_valid_addr = (unsigned long)(entry) + smap_size - 1;
first_valid_addr = (unsigned long)(entry);
if (entry->start_addr || entry->length) {
while (next_entry->start_addr || next_entry->length) {
	entry = next_entry;
	if(next_entry->start_addr)
		next_entry = (struct viva_smap_entry *)(entry->bitmap + entry->length);
	else
	{
		count = next_entry->length;
		while(count)
		{
			next_entry = (struct viva_smap_entry *)(next_entry->bitmap + 0);
			count--;
		}
		entry = next_entry;
	}
}
if (entry->length > 0) {
	/* atleast one bit is set in the sparse bitmap for this entry */
	/* we want to compute the last block's address */
	/* clearly that block must belong in the last byte of this entry */
	/* the last byte of this entry is : */

	memcpy(last_byte,(unsigned char *)(entry->bitmap + entry->length - 1),1);
	*last_byte = *last_byte ^ 0xff; /* toggle all bits */
	bit_nr = -1;
	while(bit_nr < 8) {
		old_bit_nr = bit_nr;
		bit_nr = find_next_zero_bit(last_byte,8,bit_nr+1);
	}
	if ((old_bit_nr < 0) || (old_bit_nr > 7)) {
		printk("VIVA: In viva_expand_smap: Error in sparse bitmap %ld\n",entry->start_addr);
		printk("Vamsi: bit_nr = %d, old_bit_nr = %d, last_byte = %d\n",bit_nr,old_bit_nr,*last_byte);
		return -1;
	}
	last_block_addr = entry->start_addr + (entry->length-1)*8 + old_bit_nr + 1; 
}
else {
	last_block_addr = entry->start_addr;
	if(!last_block_addr)
	{	
		entry->start_addr = new_block_addr; /* create a new entry */
		entry->length = 0; /* redundant but for added safety */
		return((unsigned long)entry + 3 - first_valid_addr);
	}
} 
if ((new_block_addr - last_block_addr) < 32) {
	bit_to_set = new_block_addr - entry->start_addr - 1;
	if ((bit_to_set/8 + 1) > 255) goto create_new_entry;
	if (((unsigned long)entry + 5 + bit_to_set/8) > last_valid_addr) {
		return -2; /* run out of space */
	}
	entry->length = bit_to_set/8 + 1;
	set_bit(bit_to_set,entry->bitmap);
	return((unsigned long)entry->bitmap + bit_to_set/8 - first_valid_addr);
}
else {
create_new_entry:
	/* else create a new entry */
	entry = (struct viva_smap_entry *)(entry->bitmap + entry->length);
	if ((unsigned long)entry > last_valid_addr) {
		return -2; /* run out of space */
	}
	entry->start_addr = new_block_addr; /* create a new entry */
	entry->length = 0; /* redundant but for added safety */
	return((unsigned long)entry + 3 - first_valid_addr);
}
}
else {
	/* No blocks in this inode */
	entry->start_addr = new_block_addr; /* create a new entry */
	entry->length = 0; /* redundant but safe */	
	return 3; /* first 4 bytes were modifed */
	/* the last bit modified was in byte number 3*/
}
}

int viva_release_smap(unsigned char *inode_sparse_bmap, int block_addr)
{

	struct viva_smap_entry *entry,*next_entry,*prev_entry;
	char last_byte[1];
	char byte[1];
	unsigned long last_block_addr,last_valid_addr,first_valid_addr;
	unsigned long prev_block_addr;
	int prev_bit_nr, last_bit_nr, bit_nr, bit_to_clear;
	int count;

	entry = next_entry = (struct viva_smap_entry *)(inode_sparse_bmap);
	last_valid_addr = (unsigned long)(entry) + VIVA_SMAP_SIZE - 1;
	first_valid_addr = (unsigned long)(entry);
	while (next_entry->start_addr || next_entry->length) 
	{
		if(!next_entry->start_addr)
			return -1;
		prev_entry = entry;
		entry = next_entry;
		next_entry =(struct viva_smap_entry *)(entry->bitmap+entry->length);
	}

	if (entry->length > 0) 
	{
		last_bit_nr = -1;
		bit_nr = -1;
		memcpy(last_byte,(unsigned char *)(entry->bitmap + entry->length-1),1);
		*last_byte = *last_byte ^ 0xff; /* toggle all bits */
		while(bit_nr < 8) {
			prev_bit_nr = last_bit_nr;
			last_bit_nr = bit_nr;
			bit_nr = find_next_zero_bit(last_byte,8,bit_nr+1);
		}
		if ((last_bit_nr < 0) || (last_bit_nr > 7)) {
			printk("VIVA: In viva_release_smap: Error in sparse bitmap\n");
			printk("Vamsi: bit_nr = %d, last_bit_nr = %d, last_byte = %d\n", bit_nr, last_bit_nr, last_byte);
			return -1;
		}
		last_block_addr = entry->start_addr+(entry->length-1)*8+last_bit_nr+1;

		bit_to_clear = last_block_addr - entry->start_addr - 1;
		clear_bit(bit_to_clear,entry->bitmap);
/*		printk("Vamsi: Cleared the bit %d (%d %d) \n",bit_to_clear,last_block_addr,entry->start_addr);*/
		if(prev_bit_nr < 0)
		{
			for(count=0; count<entry->length-1; count++)
			{   
				memcpy(byte,(unsigned char *)(entry->bitmap+count),1);
				*byte = *byte ^ 0xff; /* toggle all bits */
				bit_nr = -1;
				while(bit_nr < 8)
				{
					if(bit_nr >= 0)
						prev_bit_nr = bit_nr+ (count*8);
					bit_nr = find_next_zero_bit(byte,8,bit_nr+1);
				}
			}
			if(prev_bit_nr < 0)
			{
				entry->length = 0;
				prev_block_addr = entry->start_addr;
			}
			else
			{
				prev_block_addr = entry->start_addr + prev_bit_nr + 1;
				entry->length = prev_bit_nr/8 + 1;
			}
		}
		else
		{
			prev_block_addr=entry->start_addr+(entry->length-1)*8+prev_bit_nr+1;
/*			entry->length -= (last_block_addr/8) - (prev_block_addr/8);*/
		}
	}
	else 
	{
		if(prev_entry->start_addr != entry->start_addr)
		{
			if(!prev_entry->length)
				prev_block_addr = prev_entry->start_addr;	
			else
			{
				memcpy(last_byte,(unsigned char *)
							(prev_entry->bitmap + prev_entry->length - 1),1);
				*last_byte = *last_byte ^ 0xff; /* toggle all bits */
				bit_nr = -1;
				while(bit_nr < 8) 
				{
					last_bit_nr = bit_nr;
					bit_nr = find_next_zero_bit(last_byte,8,bit_nr+1);
				}
				if ((last_bit_nr < 0) || (last_bit_nr > 7)) 
				{
					printk("VIVA: Error in sparse bitmap\n");
					return -1;
				}
				prev_block_addr = prev_entry->start_addr + 
								  (prev_entry->length-1)*8 + last_bit_nr + 1;
			}
		}
		else
			prev_block_addr = 0;
		entry->start_addr = 0;
	} 
	return(prev_block_addr);
}
